#ifndef FLYSYSTEMSOUNDS_QT_PLAYER_H
#define FLYSYSTEMSOUNDS_QT_PLAYER_H

#include <QObject>

#include <FlySystemSoundsQt/Global>

namespace Fly {
namespace SystemSoundsQt {

//!
//! \brief The user identification is used to identify the location of the fly theme file.
//! If the stasndard identification is used, then the user will be identified by the current process.
//! In the case of non-standard authentication, the user will be searched for the owner of the current process.
//! Non-standard identification is useful when the process is started through another user,
//! such as through the `sudo` or `su` commands.
//! Default is user non-standard identification.
//! \return Is use standard user identification for fly theme file.
//!
static bool isFlyThemeUseStandardUserIdentification();

//!
//! \brief Change user identification for fly theme file.
//! \param use - Is use standard user identification for fly theme file.
//!
static void setFlyThemeUseStandardUserIdentification(bool use);

class PlayerPrivate;

//!
//! \brief The Player class provides the ability to control the playback of system sounds.
//!
class FLYSYSTEMSOUNDS_QT_EXPORT Player : public QObject
{
    Q_OBJECT

public:
    //!
    //! \brief The Mode enum:
    //! Single - Ignoring any sounds play while the current sound is playing;
    //! MultiForOneSoundFile - No ignoring other sounds: maybe parallel playing other sounds, identical sounds is ignoring when they are playing;
    //! MultiForAnySoundFile - No ignoring sounds: maybe parallel playing any sounds.
    //!
    enum class Mode {
        Single = 0,
        MultiForOneSoundFile,
        MultiForAnySoundFile,
    };
    Q_ENUM(Mode);

    //!
    //! \brief The Source enum:
    //! Current application type that calls the play functions.
    //!
    enum class Source {
        NoSoundType = 0,
        Others,
        WindowsManager,
        Notifications,
        OSD,
    };
    Q_ENUM(Source);

    //!
    //! \brief Need Q*Application instance for QEvent work.
    //! Using QCoreApplication::applicationName(), QGuiApplication::desktopFileName()
    //! and QGuiApplication::windowIcon().name() for audio application identifier.
    //!
    explicit Player(Mode mode = Mode::MultiForOneSoundFile,
                    Source source = Source::Others,
                    QObject *parent = nullptr);

    //!
    //! \brief Need Q*Application instance for QEvent work.
    //!
    explicit Player(QObject *parent = nullptr,
                    Mode mode = Mode::MultiForOneSoundFile,
                    Source source = Source::Others,
                    const QString &applicationName = QString(),
                    const QString &desktopFileName = QString(),
                    const QString &applicationIconName = QString());

    ~Player() override;

    //!
    //! \brief mode - Current operating mode for the player.
    //! This field is not changeable for an instance of the class.
    //!
    Mode mode() const;

    //!
    //! \brief sourceType - Current application type that calls the play functions.
    //! This field is not changeable for an instance of the class.
    //!
    Source sourceType() const;

    //!
    //! \brief applicationName - Current audio application identifier.
    //! Required to identify the application sounds.
    //! This field is not changeable for an instance of the class.
    //!
    QString applicationName() const;

    //!
    //! \brief desktopFileName - Current audio application identifier by desktop file.
    //! Required to identify the application sounds.
    //! This field is not changeable for an instance of the class.
    //!
    QString desktopFileName() const;

    //!
    //! \brief applicationIconName - Current audio application identifier by icon name.
    //! Required to identify the application sounds.
    //! This field is not changeable for an instance of the class.
    //!
    QString applicationIconName() const;

    //!
    //! \brief The FlySoundType enum:
    //! Standard system sound type for Fly Desktop.
    //!
    enum class FlySoundType {
        Paging = 0,
        MenuCommand,
        MenuPopup,
        Open,
        Close,
        Minimize,
        Maximize,
        RestoreDown,
        RestoreUp,
        SystemExit,
        SystemStart,
        SystemLock,
        ErrorDialog,
        ForceTrayMessage,
        OSDNotify,
    };
    Q_ENUM(FlySoundType);

    //!
    //! \return Is the fly sound theme currently enabled
    //!
    static bool isFlySoundsEnabled();

    //!
    //! \param soundType - fly sound type.
    //! \return Is the fly sound theme currently enabled by sound type
    //!
    bool isFlySoundTypeEnabled(FlySoundType soundType) const;

    //!
    //! \param soundType - fly sound type.
    //! \return Sound name by fly
    //!
    static QString flySoundTypeName(FlySoundType soundType);

    //!
    //! \param flySoundTypeName - sound name by fly
    //! \param ok - bool pointer for validating flySoundTypeName
    //! \return Fly sound type
    //!
    static FlySoundType flySoundTypeByName(const QString &flySoundTypeName, bool *ok = nullptr);

    //!
    //! \brief Changes whether the cache is enabled for all types of fly system sounds.
    //! \param isEnabled - is cache enabled.
    //!
    void setFlySystemSoundPathCacheEnabled(bool isEnabled);

    //!
    //! \brief Cache for all types of fly system sounds.
    //! \return is cache enabled.
    //!
    bool isFlySystemSoundPathCacheEnabled() const;

    typedef QMap<Player::FlySoundType, QString> FlySoundTypePathsCache;

    //!
    //! \return Map of fly system sounds paths.
    //!
    FlySoundTypePathsCache flySystemSoundPathCache() const;

    //!
    //! \brief When using a flySystemSoundPathCache, the information will be taken from the flySystemSoundPathCache.
    //! \param soundType - fly sound type.
    //! \return Sound file path by fly
    //!
    QString flySoundFilePathByType(Player::FlySoundType soundType) const;

    //!
    //! \return Is anything playing sounds right now
    //!
    bool isPlaying() const;

    //!
    //! \param soundType - fly sound type.
    //! \return Is the type of sound currently playing
    //!
    bool isPlayingByType(FlySoundType soundType) const;

    //!
    //! \param filePath - sound file name.
    //! \return Is the audio file name currently playing
    //!
    bool isPlayingByName(const QString &name);

    //!
    //! \param filePath - sound file path.
    //! \return Is the audio file currently playing
    //!
    bool isPlayingByPath(const QString &filePath);

    //!
    //! \param flySoundName - sound name by fly
    //! \return Is the type of fly sound currently playing
    //!
    bool isPlayingByFlySoundName(const QString &flySoundName);

    //!
    //! \brief The PlayExitCode enum:
    //! Success - Successful start of audio play;
    //! ModeRestriction - The sound did not play due to the current mode limitation;
    //! Error - Standard file start play error, a description of the error will be printed to stderr;
    //!
    enum class PlayExitCode {
        Success = 0,
        ModeRestriction,
        Error
    };
    Q_ENUM(PlayExitCode);

    //!
    //! \brief The ServerSoundCachePolicy enum:
    //! A special property that can be used to control the automatic sound
    //! caching of sounds in the sound server. This is useful for very
    //! frequently used sound events such as those used for input feedback.
    //! Never - may be used for sounds that should never be cached;
    //! Volatile - may be used for cacheing sounds in the sound server temporarily;
    //! Permanent - will cause this sample to be cached in the server permanently.
    //!
    enum class ServerSoundCachePolicy {
        Never = 0,
        Volatile,
        Permanent,
    };
    Q_ENUM(ServerSoundCachePolicy);

public slots:
    //!
    //! \brief Sound launch by type.
    //! \param soundType - fly sound type.
    //! \param fileSoundCachePolicy - cache policy for current sound file.
    //! \return Exit code for start play operation
    //!
    PlayExitCode playByType(FlySoundType soundType,
                            ServerSoundCachePolicy fileSoundCachePolicy = ServerSoundCachePolicy::Volatile);

    //!
    //! \brief Sound launch by file name.
    //! \param name - sound file name, сan be used with directory separators.
    //! \param fileSoundCachePolicy - cache policy for current sound file.
    //! \return Exit code for start play operation
    //!
    PlayExitCode playByName(const QString &name,
                            ServerSoundCachePolicy fileSoundCachePolicy = ServerSoundCachePolicy::Volatile);

    //!
    //! \brief Sound launch by file path.
    //! \param filePath - sound file path.
    //! \param fileSoundCachePolicy - cache policy for current sound file.
    //! \return Exit code for start play operation
    //!
    PlayExitCode playByPath(const QString &filePath,
                            ServerSoundCachePolicy fileSoundCachePolicy = ServerSoundCachePolicy::Volatile);

    //!
    //! \brief Sound launch by fly sound name.
    //! \param flySoundName - sound name by fly.
    //! \param fileSoundCachePolicy - cache policy for current sound file.
    //! \return Exit code for start play operation
    //!
    PlayExitCode playByFlySoundName(const QString &flySoundName,
                                    ServerSoundCachePolicy fileSoundCachePolicy = ServerSoundCachePolicy::Volatile);

    //!
    //! \brief Cancel play immediately by fly sound type.
    //! \param soundType - fly sound name.
    //!
    void cancelByType(FlySoundType soundType);

    //!
    //! \brief Cancel play immediately by sound name.
    //! \param flySoundName - sound name, сan be used with directory separators.
    //!
    void cancelByName(const QString &name);

    //!
    //! \brief Cancel play immediately by file path.
    //! \param filePath - sound file path.
    //!
    void cancelByPath(const QString &filePath);

    //!
    //! \brief Cancel play immediately by fly sound name.
    //! \param flySoundName - sound name by fly.
    //!
    void cancelByFlySoundName(const QString &flySoundName);

signals:
    //!
    //! \brief Signal for play finish.
    //! \param filePath - sound file path.
    //! \param isSuccess - successful completion of audio play.
    //!
    void finished(const QString &filePath, bool isSuccess);

    //!
    //! \brief Signal on fly system sound path cache change.
    //! \param flySystemSoundPathCache - new fly system sound path cache.
    //!
    void flySystemSoundPathCacheChanged(const FlySoundTypePathsCache &flySystemSoundPathCache);

private slots:
    void onSoundThemeConfigFileChanged(const QString &path);

    void updateFlySystemSoundPathCache();

private:
    Q_DISABLE_COPY(Player);

#if QT_VERSION > QT_VERSION_CHECK(5, 13, 0)
    Q_DISABLE_MOVE(Player);
#else
    Player(Player &&) = delete;
    Player &operator=(Player &&) = delete;
#endif // Qt > 5.13.0

    PlayerPrivate *d;

    void init(Mode mode,
              Source source,
              const QString &applicationName,
              const QString &desktopFileName,
              const QString &applicationIconName);
};

inline bool isSuccess(Player::PlayExitCode playExitCode)
{
    return playExitCode == Player::PlayExitCode::Success;
}

} } // namespace Fly::SystemSoundsQt

#endif // FLYSYSTEMSOUNDS_QT_PLAYER_H
